package edu.gmu.cs475;

import java.io.IOException;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.function.BiConsumer;

public abstract class AbstractKeyValueLoader {
	public AbstractKeyValueStore kvStore;
	ExecutorService executor;
	BiConsumer<String, String> listener;

	public AbstractKeyValueLoader() {
		this.kvStore = new KeyValueStore();
		executor = Executors.newFixedThreadPool(4);
	}

	/**
	 * Given a single line, load it into the kvStore.
	 * 
	 * The line is guaranteed to be of the format key=newValue, where "key" may not
	 * contain an equals character ('='), and neither key nor value are null.
	 * 
	 * The process for storing a line key=newValue is:
	 * 1. Check to see if the key exists
	 * 2. If the key does not exist, then add this key/value pair (as a string)
	 * 3. If the key does exist, and it is a list, then add this value to that key's list
	 * 4. If the key does exist, and its value is a string [oldValue], then replace the string with a list,
	 *      containing the initial elements [oldValue, newValue]. The list must be created by calling newList.
	 *
	 * @param line
	 */
	protected abstract void loadLine(String line);

	/**
	 * Loads all of the lines in the given file into the kvStore
	 *
	 * Must load lines in parallel. You must perform this parallelization by
	 * submitting tasks to the executor object -- one per line in the inputFile
	 *
	 * @param inputFile file to load in
	 * @throws IOException if any underlying error occurs when reading the file
	 */
	protected abstract void loadFile(Path inputFile) throws IOException;

	/**
	 * Creates a new list to store in the key value store, possibly with some initial value
	 *
	 * @param forKey        the key that will be storing this list
	 * @param initialValues any initial values to put in this list
	 * @return the new list
	 */
	protected final List<String> newList(String forKey, String... initialValues) {
		ArrayList<String> ret = new ArrayList<String>() {
			@Override
			public boolean add(String o) {
				if (listener != null)
					listener.accept(forKey, o);
				return super.add(o);
			}
		};
		if (initialValues != null)
			for (String s : initialValues)
				ret.add(s);
		return ret;
	}

	public final void setListener(BiConsumer<String, String> addToListListener) {
		listener = addToListListener;
	}
}
