package edu.gmu.cs475;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Set;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.function.BiConsumer;

/**
 * Do not modify this file.
 */
public abstract class AbstractKeyValueStore {

	/**
	 * You must use this lock when accessing
	 */
	protected final Lock bigLock = new ReentrantLock();
	private final HashMap<String, Object> map = new HashMap<>();

	/**
	 * Retrieve an element from this key value store
	 *
	 * @param key the key to retrieve
	 * @return The value mapped to this key, if one exists, otherwise null
	 * @throws NullPointerException if key is null
	 */
	public abstract Object get(String key);

	/**
	 * Returns the type of element that is stored at this key
	 *
	 * @param key the key to inquire about
	 * @return A ValueType indicating the kind of value mapped to this key, if one exists, otherwise null
	 * @throws NullPointerException if key is null
	 */
	public abstract ValueType getType(String key);

	/**
	 * Lists all of the keys that are currently known to this key-value store
	 *
	 * @return A set containing all currently valid keys
	 */
	public abstract Set<String> listKeys();

	/**
	 * Sets a key to be the given value
	 *
	 * @param key   key to set
	 * @param value value to store
	 * @throws NullPointerException     if key or value is null
	 * @throws IllegalArgumentException if value is not a String or List.
	 */
	public abstract void set(String key, Object value);

	/**
	 * Removes this key and any values stored at it from the map
	 *
	 * @param key key to remove
	 * @return true if the key was successfully deleted, or false if it did not exist
	 * @throws NullPointerException if key is null
	 */
	public abstract boolean remove(String key);

	/**
	 * Appends the given item to the list stored at the given key
	 *
	 * @param key     key that contains the list
	 * @param newItem item to append to the list at the given key
	 * @throws NullPointerException     if key is null
	 * @throws IllegalArgumentException if key doesn't exist, or key does not represent a list
	 */
	public abstract void addToList(String key, String newItem) throws IllegalArgumentException;

	/**
	 * Deletes at most one occurrence of itemToDelete from the list stored at the given key
	 *
	 * @param key          key containing the list
	 * @param itemToDelete item to delete from the list
	 * @throws IllegalArgumentException if key doesn't exist, or key does not represent a list
	 * @throws NullPointerException     if key is null
	 * @returns true if the item was deleted from the list (false if itemToDelete didn't exist in that list)
	 */
	public abstract boolean deleteFromList(String key, String itemToDelete) throws IllegalArgumentException;


	private final LinkedList<BiConsumer<String, String>> listeners = new LinkedList<>();

	public final void addListener(BiConsumer<String, String> listener) {
		listeners.add(listener);
	}

	/**
	 * Retrieve an item from the underlying store - you must call this from your KeyValueStore
	 * <p>
	 * You must hold the bigLock in order to call this method
	 *
	 * @param key key to retrieve
	 * @return The value stored at the given key, or null if none exists
	 * @throws NullPointerException if key is null
	 */
	protected final Object _get(String key) {
		for (BiConsumer<String, String> l : listeners)
			l.accept("_get", key);
		if (key == null)
			throw new NullPointerException();
		return map.get(key);
	}

	/**
	 * Add an item to the underlying store - you must call this from your KeyValueStore
	 * <p>
	 * You must hold the bigLock in order to call this method
	 *
	 * @param key   key to set
	 * @param value value to store
	 * @throws NullPointerException if key or value is null
	 */
	protected final void _set(String key, Object value) {
		for (BiConsumer<String, String> l : listeners)
			l.accept("_set", key);
		if (key == null || value == null) throw new NullPointerException();
		map.put(key, value);
	}

	/**
	 * Remove an item from the underlying store - you must call this from your KeyValueStore
	 * <p>
	 * You must hold the bigLock in order to call this method
	 *
	 * @param key key to remove
	 * @return true if the value was removed, false if not
	 * @throws NullPointerException if key is null
	 */
	protected final boolean _remove(String key) {
		for (BiConsumer<String, String> l : listeners)
			l.accept("_remove", key);
		if (key == null)
			throw new NullPointerException();
		return map.remove(key) != null;
	}

	/**
	 * Enumerates all of the keys currently in the map
	 * <p>
	 * You must hold the bigLock in order to call this method
	 *
	 * @return Set containing all currently valid keys
	 */
	protected final Set<String> _listKeys() {
		for (BiConsumer<String, String> l : listeners)
			l.accept("_listKeys", null);
		return new HashSet<>(map.keySet());
	}

	@Override
	public String toString() {
		return "KeyValueStore{" +
				"map=" + map +
				'}';
	}

	public enum ValueType {
		STRING,
		LIST
	}
}
